from ..formation import _CONTEXT, _REQ_ID, _REQ_PARENT_ID, _SESSION, _UID
import os
from six.moves import _thread as thread
from toolz.curried import get_in


def get_context(
    request_id=None,
    request_parent_id=None,
    namespace="service",
    env="local",
    sha="dev",
    version="0.0.1",
    scope="service",
    uid=None,
    getpid=os.getpid,
    gettid=thread.get_ident,
):
    pid = getpid()
    tid = gettid()
    return {
        "v": version,
        "sha": sha,
        "env": env,
        "pid": pid,
        "tid": tid,
        "uid": uid,
        "scope": scope,
        "ns": namespace,
        "rid": request_id,
        "rid_p": request_parent_id,
    }


def call_context_fnc(ctx, context_fn, namespace, scope, env, sha, version, getpid, gettid):
    request_id = ctx.get(_REQ_ID, None)
    request_parent_id = ctx.get(_REQ_PARENT_ID, None)
    uid = get_in([_SESSION, _UID], None)

    ctx[_CONTEXT] = context_fn(
        env=env,
        sha=sha,
        namespace=namespace,
        version=version,
        request_id=request_id,
        request_parent_id=request_parent_id,
        scope=scope,
        uid=uid,
        getpid=getpid,
        gettid=gettid,
    )
    return ctx


def context(
    context_fn=get_context,
    namespace="service",
    scope="all",
    env="local",
    sha="dev",
    version="0.01",
    getpid=os.getpid,
    gettid=thread.get_ident,
):
    def context_middleware(ctx, call):
        ctx = call_context_fnc(
            ctx, context_fn, namespace, scope, env, sha, version, getpid, gettid
        )
        ctx = call(ctx)
        return ctx

    return context_middleware


def async_context(
    context_fn=get_context,
    namespace="service",
    scope="all",
    env="local",
    sha="dev",
    version="0.01",
    getpid=os.getpid,
    gettid=thread.get_ident,
):
    async def context_middleware(ctx, call):
        ctx = call_context_fnc(
            ctx, context_fn, namespace, scope, env, sha, version, getpid, gettid
        )
        ctx = await call(ctx)
        return ctx

    return context_middleware



from ..formation import _REQ_HTTP


def update_req_for_timeout(ctx, timeout) -> None:
    req = ctx.get(_REQ_HTTP)
    if timeout:
        req.timeout = timeout


def timeout(timeout=None):
    def timeout_middleware(ctx, call):
        update_req_for_timeout(ctx, timeout)
        return call(ctx)

    return timeout_middleware


def async_timeout(timeout=None):
    async def timeout_middleware(ctx, call):
        update_req_for_timeout(ctx, timeout)
        return await call(ctx)

    return timeout_middleware



import datetime
from abc import abstractmethod, ABC
from typing import Type
from toolz import reduce
from toolz.curried import keyfilter, reduce
from urllib.parse import urljoin
from attr import attrib, attrs
from functools import partial

_REQ_HTTP = "fmtn.req.http"
_RES_HTTP = "fmtn.res.http"
_CONTEXT = "fmtn.context"
_SESSION = "fmtn.session"
_RETRY = "fmtn.retry"
_REQ_ID = "req.id"
_UID = "uid"
_REQ_PARENT_ID = "req.parent.id"
_REQ_DURATION = "req.duration_us"
_RES_DEFAULT_TIME_OUT = 30
_RETRY_IN_BETWEEN_CALL_SLEEP = 10


class BaseSender(ABC):
    def __init__(self, middleware, base_uri, default_response_as):
        self.middleware = middleware
        self.base_uri = base_uri
        self.default_response_as = default_response_as

    def get(self, path, **kwargs):
        return self.send("get", path, **kwargs)

    def post(self, path, **kwargs):
        return self.send("post", path, **kwargs)

    def put(self, path, **kwargs):
        return self.send("put", path, **kwargs)

    def patch(self, path, **kwargs):
        return self.send("patch", path, **kwargs)

    def delete(self, path, **kwargs):
        return self.send("delete", path, **kwargs)

    @abstractmethod
    def send(self, method, url, session_context=None, params=None, response_as=None, **kwargs):
        pass


def client_decorator(cls, sender_class: Type[BaseSender], session_class: Type[BaseSender] = None):
    original_init = cls.__init__

    def now_iso(self):
        return datetime.datetime.utcnow().isoformat()

    def path(self, p):
        return urljoin(self.base_uri, p)

    def init(self, *args, **kwargs):
        original_init(self, *args, **kwargs)
        self.base_uri = kwargs.get(
            "base_uri", getattr(self.__class__, "base_uri", "http://localhost")
        )
        response_as = kwargs.get(
            "response_as", getattr(self.__class__, "response_as", None)
        )
        middleware = kwargs.get(
            "middleware", getattr(self.__class__, "middleware", [])
        ) or []
        sender_class_param = getattr(self.__class__, "sender_class", None) or sender_class
        self.request = sender_class_param(
            middleware=middleware, base_uri=self.base_uri, default_response_as=response_as
        )
        if session_class:
            self.session = partial(session_class,
                middleware=middleware, base_uri=self.base_uri, default_response_as=response_as
            )

    cls.path = path
    cls.now_iso = now_iso
    cls.__init__ = init
    return cls


@attrs
class FormationHttpRequest:
    url = attrib()
    method = attrib(default="get")

    # these two are very stabby. a single default instance is shared among all attrs
    # objects. to assign new keys, update immutably -- use merge and re-assign
    headers = attrib(factory=dict)
    cookies = attrib(factory=dict)
    params = attrib(factory=dict)

    auth = attrib(default=None)
    data = attrib(default=None)
    json = attrib(default=None)
    files = attrib(default=None)
    timeout = attrib(default=None)
    allow_redirects = attrib(default=True)


def wrap(call, middleware=None):
    if middleware is None:
        middleware = []
    return reduce(
        lambda acc, m: lambda ctx: m(ctx, acc),
        reversed(middleware),
        lambda ctx: call(ctx),
    )


def params_filter(p):
    return p.startswith(":")


def not_params_filter(p):
    return not params_filter(p)


def apply_params(url, params):
    route_params = keyfilter(params_filter, params)
    return (
        reduce(lambda acc, kv: acc.replace(kv[0], kv[1]), route_params.items(), url),
        keyfilter(not_params_filter, params),
    )


def get_response(ctx):
    return ctx.get(_RES_HTTP, None)



from .breaker import (
    circuit_breaker,
    trigger_breaker_if,
    BreakerTriggerException,
)  # noqa

from .iobreaker import (
    async_circuit_breaker,
    async_trigger_breaker_if,
    AsyncBreakerTriggerException,
)  # noqa

from .context import context, async_context
from .context_logger import context_logger, async_context_logger  # noqa
from .logger import request_logger, async_request_logger
from .request_duration import request_duration, async_request_duration
from .request_id import request_id, async_request_id
from .retry import retry, async_retry  # noqa
from .ua import ua, async_ua  # noqa
from .accept import accept, async_accept  # noqa
from .timeout import timeout, async_timeout  # noqa


def default_stack(logger):
    return [request_id(), context(), request_duration(), request_logger(logger)]



from toolz import merge
from ..formation import _REQ_HTTP


def update_header_for_ua(ctx, user_agent: str) -> None:
    req = ctx.get(_REQ_HTTP)
    req.headers = merge(req.headers, {"User-Agent": user_agent})


def ua(user_agent="formation/1.0.0"):
    def ua_middleware(ctx, call):
        update_header_for_ua(ctx, user_agent)
        return call(ctx)

    return ua_middleware


def async_ua(user_agent="formation/1.0.0"):
    async def ua_middleware(ctx, call):
        update_header_for_ua(ctx, user_agent)
        return await call(ctx)

    return ua_middleware



from urllib.parse import urljoin
from attr import attrib, attrs
import xmltodict
from aiohttp import ClientSession, ClientTimeout, BasicAuth

from .formation import (
    FormationHttpRequest, apply_params, client_decorator, get_response, wrap,
    _REQ_HTTP,
    _RES_HTTP,
    _SESSION,
    _RES_DEFAULT_TIME_OUT,
    _REQ_DURATION, BaseSender
)


__all__ = [
    "async_client",
    "async_json_response",
    "async_xmltodict_response",
    "async_text_response",
    "async_raw_response",
]


def async_client(cls):
    return client_decorator(cls, Sender)


@attrs
class FormationAIOHttpRequest(FormationHttpRequest):
    content_resolver = attrib(default=None)


def get_response_result(ctx):
    response = get_response(ctx)
    if response is None:
        return None, None, None
    return response, response.status, response.headers


async def async_raw_response(response):
    if response is not None:
        response_data = await response.read()
        response.parsed_content = response_data
    return response


async def async_json_response(response):
    if response is not None:
        response_data = await response.json(content_type=None)
        response.parsed_content = response_data
    return response


async def async_xmltodict_response(response):
    if response is not None:
        response_txt = await response.text()
        response_data = xmltodict.parse(response_txt)
        response.parsed_content = response_data
    return response


async def async_text_response(response):
    if response is not None:
        response_data = await response.text()
        response.parsed_content = response_data
    return response


class Sender(BaseSender):
    def __init__(self, middleware=None, base_uri=None, default_response_as=None):
        super().__init__(middleware, base_uri, default_response_as or async_raw_response)

    async def send(self, method, url, session_context=None, params=None, response_as=None, **kwargs):
        params = params or {}
        params = params if isinstance(params, dict) else params.to_dict()
        (url, params) = apply_params(url, params)
        request = FormationAIOHttpRequest(
            url=urljoin(self.base_uri, url),
            method=method,
            params=params,
            content_resolver=response_as or self.default_response_as,
            **kwargs
        )
        return await self.send_request(request, session_context)

    async def send_request(self, request: FormationAIOHttpRequest, session_context=None):
        ctx = {_REQ_HTTP: request}
        if session_context:
            ctx[_SESSION] = session_context
        ctx = await wrap(requests_adapter, middleware=self.middleware)(ctx)
        return get_response_result(ctx)


async def requests_adapter(ctx):
    req = ctx[_REQ_HTTP]
    auth = None
    if req.auth:
        auth = BasicAuth(req.login, req.password, encoding="latin1")
    timeout = (
        ClientTimeout(total=req.timeout)
        if req.timeout
        else ClientTimeout(total=_RES_DEFAULT_TIME_OUT)
    )
    async with ClientSession(
        headers=req.headers, cookies=req.cookies, timeout=timeout
    ) as session:
        method = getattr(session, req.method.lower())
        async with method(
            url=req.url, params=req.params, data=req.data, auth=auth, cookies=req.cookies, json=req.json
        ) as response:
            async with response:
                response = await req.content_resolver(response)
                # Compatibility with original response objects
                response.request = req
                response.status_code = response.status
                response.elapsed = ResponseElapsed(ctx)
                ctx[_RES_HTTP] = response

            return ctx


class ResponseElapsed:
    def __init__(self, ctx):
        self.ctx = ctx

    def total_seconds(self):
        return self.ctx.get(_REQ_DURATION) / 10e6



import pybreaker
from ..formation import _CONTEXT, _RES_HTTP


class BreakerTriggerException(Exception):
    pass


def breaker_logger(logger):
    class LogListener(pybreaker.CircuitBreakerListener):
        def state_change(self, cb, old_state, new_state):
            logger.warn(
                "circuitbreaker.state_changed",
                name=cb.name,
                old_state=old_state.name,
                new_state=new_state.name,
            )

    return LogListener()


def trigger_breaker_if(trigger):
    def trigger_breaker_middleware(ctx, call):
        ctx = call(ctx)
        if trigger(ctx.get(_RES_HTTP)):
            raise BreakerTriggerException

    return trigger_breaker_middleware


def circuit_breaker(
    logger, name, fail_max=5, reset_timeout=60, state_storage=None, exclude=None
):
    if exclude is None:
        exclude = []
    breaker = pybreaker.CircuitBreaker(
        name=name,
        listeners=[breaker_logger(logger)],
        exclude=exclude,
        fail_max=fail_max,
        reset_timeout=reset_timeout,
        state_storage=state_storage,
    )

    def circuit_breaker_middleware(ctx, call):
        context = ctx.get(_CONTEXT, {})
        log = logger.bind(**context)

        if breaker.current_state == "open":
            log.info("circuitbreaker.open", name=breaker.name)

        call = breaker(call)

        try:
            ctx = call(ctx)
            return ctx
        except pybreaker.CircuitBreakerError:
            return ctx

    return circuit_breaker_middleware



from ..formation import _RETRY, _RETRY_IN_BETWEEN_CALL_SLEEP
from time import sleep
import asyncio


def retry(max_retries=3, retry_in_between_call_sleep=_RETRY_IN_BETWEEN_CALL_SLEEP):
    def retry_middleware(ctx, call):
        try:
            res = call(ctx)
            return res
        except Exception as ex:
            retries = ctx.get(_RETRY, 0)
            if retries >= max_retries - 1:
                raise ex
            ctx[_RETRY] = 1 + retries
            # TODO exponential backoff
            sleep(retry_in_between_call_sleep)
            res = retry_middleware(ctx, call)
            return res

    return retry_middleware


def async_retry(
    max_retries=3, retry_in_between_call_sleep=_RETRY_IN_BETWEEN_CALL_SLEEP
):
    async def retry_middleware(ctx, call):
        try:
            res = await call(ctx)
            return res
        except Exception as ex:
            retries = ctx.get(_RETRY, 0)
            if retries >= max_retries - 1:
                raise ex
            ctx[_RETRY] = 1 + retries
            # TODO exponential backoff
            await asyncio.sleep(retry_in_between_call_sleep)
            res = await retry_middleware(ctx, call)
            return res

    return retry_middleware



from ..formation import _REQ_HTTP, _RES_HTTP, _CONTEXT, _REQ_DURATION


def bind_logger(ctx, logger):
    context = ctx.get(_CONTEXT, {})
    logger.bind(**context).info("context")


def context_logger(logger):
    def context_logger_middleware(ctx, next):
        bind_logger(ctx, logger)
        ctx = next(ctx)
        return ctx

    return context_logger_middleware


def async_context_logger(logger):
    async def context_logger_middleware(ctx, next):
        bind_logger(ctx, logger)
        ctx = await next(ctx)
        return ctx

    return context_logger_middleware



from ..formation import (
    _REQ_HTTP,
    _RES_HTTP,
    _CONTEXT,
    _REQ_DURATION,
)
from toolz.curried import valfilter


def request_logger(logger):
    no_nones = valfilter(lambda x: x)

    def request_logger_middleware(ctx, next):
        req = ctx[_REQ_HTTP]
        context = ctx.get(_CONTEXT, {})
        msg = "request.http"

        log = logger.bind(**context)
        log.info(msg, url=req.url, method=req.method, params=no_nones(req.params))
        log.debug(msg, headers=req.headers)

        ctx = next(ctx)

        res = ctx[_RES_HTTP]

        msg = "response.http"
        log.info(
            msg,
            url=res.request.url,
            status=res.status_code,
            method=res.request.method,
            elapsed=res.elapsed,
            size=len(res.content),
            duration_us=ctx.get(_REQ_DURATION, None),
        )
        log.debug(msg, headers=res.headers)
        return ctx

    return request_logger_middleware


def async_request_logger(logger):
    no_nones = valfilter(lambda x: x)

    async def request_logger_middleware(ctx, next):
        req = ctx[_REQ_HTTP]
        context = ctx.get(_CONTEXT, {})
        msg = "before request.http"

        log = await logger.bind(**context)
        await log.info(msg, url=req.url, method=req.method, params=no_nones(req.params))
        await log.debug(msg, headers=req.headers)

        ctx = await next(ctx)

        res = ctx[_RES_HTTP]

        msg = "after response.http"
        await log.info(
            msg,
            url=res.request.url,
            status=res.status_code,
            method=res.request.method,
            elapsed=res.elapsed,
            size=len(res.parsed_content if hasattr(res, "parsed_content") else res.content),
            duration_us=ctx.get(_REQ_DURATION, None),
        )
        await log.debug(msg, headers=res.headers)
        return ctx

    return request_logger_middleware



from toolz import merge
from ..formation import _REQ_HTTP, _REQ_ID
from uuid import uuid4


def update_req_for_req_id(ctx, key: str, idgen) -> None:
    headers = ctx[_REQ_HTTP].headers
    ctx[_REQ_HTTP].headers = merge(headers, {key: headers.get(key, str(idgen()))})
    ctx[_REQ_ID] = ctx[_REQ_HTTP].headers[key]


def request_id(key="x-request-id", idgen=uuid4):
    def request_id_middleware(ctx, next):
        update_req_for_req_id(ctx, key, idgen)
        ctx = next(ctx)
        return ctx

    return request_id_middleware


def async_request_id(key="x-request-id", idgen=uuid4):
    async def request_id_middleware(ctx, next):
        update_req_for_req_id(ctx, key, idgen)
        ctx = await next(ctx)
        return ctx

    return request_id_middleware



from aiobreaker import CircuitBreaker, CircuitBreakerError
from aiobreaker.listener import CircuitBreakerListener
from aiobreaker.state import CircuitBreakerState
from ..formation import _CONTEXT, _RES_HTTP
from datetime import timedelta


class AsyncBreakerTriggerException(Exception):
    pass


def breaker_logger(logger):
    class LogListener(CircuitBreakerListener):
        async def state_change(self, cb, old_state, new_state):
            print(f"old_state = {old_state}, new_state = {new_state}")
            await logger.info(
                "circuitbreaker.state_changed",
                name=cb.name,
                old_state=old_state.name,
                new_state=new_state.name,
            )

    return LogListener()


def async_trigger_breaker_if(trigger):
    async def trigger_breaker_middleware(ctx, call):
        ctx = await call(ctx)
        if trigger(ctx.get(_RES_HTTP)):
            raise AsyncBreakerTriggerException
        return ctx

    return trigger_breaker_middleware


def async_circuit_breaker(
    logger, name, fail_max=3, reset_timeout=60, state_storage=None, exclude=None
):
    if exclude is None:
        exclude = []
    breaker = CircuitBreaker(
        name=name,
        listeners=[breaker_logger(logger)],
        exclude=exclude,
        fail_max=fail_max,
        timeout_duration=timedelta(
            days=0, hours=0, minutes=0, seconds=reset_timeout, milliseconds=0
        ),
        state_storage=state_storage,
    )

    async def circuit_breaker_middleware(ctx, call):
        context = ctx.get(_CONTEXT, {})
        log = await logger.bind(**context)

        if breaker.current_state == CircuitBreakerState.OPEN:
            print(f"breaker.name = {breaker.name}")
            await log.info("circuitbreaker.open", name=breaker.name)

        call = breaker(call)

        try:
            ctx = await call(ctx)
            return ctx
        except CircuitBreakerError:
            return ctx

    return circuit_breaker_middleware



from toolz import merge
from ..formation import _REQ_HTTP


def update_req_for_mime_type(ctx, mime_type):
    req = ctx.get(_REQ_HTTP)
    req.headers = merge(req.headers, {"Content-Type": mime_type})


def accept(mime_type):
    def accept_middleware(ctx, call):
        update_req_for_mime_type(ctx, mime_type)
        return call(ctx)

    return accept_middleware


def async_accept(mime_type):
    async def accept_middleware(ctx, call):
        update_req_for_mime_type(ctx, mime_type)
        return await call(ctx)

    return accept_middleware



from cytoolz.curried import get_in, merge, update_in, reduce, pipe, map, filter
from attr import attrib, attrs, fields, asdict


def serde(cls=None, from_key="from", to_key="to"):
    def serde_with_class(cls):
        from_fields = list(
            map(lambda a: (a, get_in([from_key], a.metadata, [a.name])), fields(cls))
        )

        to_fields = pipe(
            fields(cls),
            map(lambda a: (a, get_in([to_key], a.metadata))),
            filter(lambda f: f[1]),
            list,
        )

        def from_dict(d):
            return cls(
                **dict(
                    map(
                        lambda f: (f[0].name, get_in(f[1], d, f[0].default)),
                        from_fields,
                    )
                )
            )

        def to_dict(self):
            d = asdict(self)
            if not to_fields:
                return d

            return reduce(
                lambda acc, f: update_in(acc, f[1], lambda _: d[f[0].name]),
                to_fields,
                {},
            )

        cls.from_dict = staticmethod(from_dict)
        cls.to_dict = to_dict
        return cls

    if cls:
        return serde_with_class(cls)
    else:
        return serde_with_class


import datetime
import math
import asyncio

from ..formation import _REQ_DURATION

SEC_TO_MICROSEC = 1e6


def request_duration(now=datetime.datetime.now):
    def request_duration_middleware(ctx, next):
        start = now()
        ctx = next(ctx)
        end = now() - start
        ctx[_REQ_DURATION] = math.floor(end.total_seconds() * SEC_TO_MICROSEC)
        return ctx

    return request_duration_middleware


def async_request_duration():
    async def request_duration_middleware(ctx, next):
        start = asyncio.get_event_loop().time()
        ctx = await next(ctx)
        end = asyncio.get_event_loop().time() - start
        ctx[_REQ_DURATION] = math.floor(end * SEC_TO_MICROSEC)
        return ctx

    return request_duration_middleware



from abc import ABC, abstractmethod
from collections.abc import Iterable
import requests
from urllib.parse import urljoin
from .formation import FormationHttpRequest, apply_params, client_decorator, get_response, wrap, _REQ_HTTP, _RES_HTTP, \
    _SESSION, BaseSender
from lxml import html
import xmltodict

__all__ = [
    "client",
    "raw_response",
    "json_response",
    "xmltodict_response",
    "html_response",
    "text_response",
]


def client(cls):
    return client_decorator(cls, Sender, SessionSender)


def _raw_response(ctx):
    res = get_response(ctx)
    if res is None:
        return None, None, None
    return res, res.status_code, res.headers


def raw_response(ctx):
    return _raw_response(ctx)


def json_response(ctx):
    res = get_response(ctx)
    if res is None:
        return None, None, None
    return res.json(), res.status_code, res.headers


def xmltodict_response(ctx):
    res = get_response(ctx)
    if res is None:
        return None, None, None
    return xmltodict.parse(res.text), res.status_code, res.headers


def html_response(ctx):
    res = get_response(ctx)
    if res is None:
        return None, None, None
    return html.fromstring(res.content), res.status_code, res.headers


def text_response(ctx):
    res = get_response(ctx)
    if res is None:
        return None, None, None
    return res.text, res.status_code, res.headers


class BaseRequestsSender(BaseSender, ABC):
    def __init__(self, middleware=None, base_uri=None, default_response_as=None):
        super().__init__(middleware, base_uri, default_response_as or _raw_response)

    def send(self, method, url, session_context=None, params=None, response_as=None, **kwargs):
        params = params or {}
        params = params if isinstance(params, Iterable) else params.to_dict()
        (url, params) = apply_params(url, params)
        request = FormationHttpRequest(
            url=urljoin(self.base_uri, url), method=method, params=params, **kwargs
        )
        return self.send_request(request, session_context, response_as)

    def send_request(self, request: FormationHttpRequest, session_context=None, response_as=None):
        ctx = {_REQ_HTTP: request}
        if session_context:
            ctx[_SESSION] = session_context
        ctx = wrap(self._get_adapter, middleware=self.middleware)(ctx)
        resolved_response_as = response_as or self.default_response_as
        return resolved_response_as(ctx)

    @abstractmethod
    def _get_adapter(self, ctx):
        pass


class Sender(BaseRequestsSender):
    def _get_adapter(self, ctx):
        return _requests_adapter(ctx, requests)


class SessionSender(BaseRequestsSender):
    def __enter__(self):
        self.session = requests.Session()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()

    def _get_adapter(self, ctx):
        return _requests_adapter(ctx, self.session)


def _requests_adapter(ctx, requests_obj):
    req = ctx[_REQ_HTTP]
    method = getattr(requests_obj, req.method.lower())
    ctx[_RES_HTTP] = method(
        req.url,
        headers=req.headers,
        cookies=req.cookies,
        params=req.params,
        auth=req.auth,
        data=req.data,
        json=req.json,
        files=req.files,
        timeout=req.timeout,
        allow_redirects=req.allow_redirects,
    )
    return ctx
